from django.test import TestCase
from ievv_auth.ievv_jwt.settings import DEFAULT_SETTINGS
from ievv_auth.ievv_jwt.utils.load_settings import load_settings


class TestLoadSettings(TestCase):
    def test_load_settings_sanity(self):
        settings = load_settings(backend_type='default')
        self.assertDictEqual(
            settings,
            DEFAULT_SETTINGS
        )

    def test_load_settings_with_default_settings_override(self):
        with self.settings(IEVV_JWT={
            'default': {
                'ISSUER': 'ievv issuer',
            }
        }):
            settings = load_settings(backend_type='default')
            settings2 = DEFAULT_SETTINGS
            settings2['ISSUER'] = 'ievv issuer'
            self.assertDictEqual(
                settings,
                settings2
            )

    def test_load_default_settings_with_non_existing_jwt_backend_type(self):
        settings = load_settings(backend_type='non-existing-backend')
        self.assertDictEqual(
            settings,
            DEFAULT_SETTINGS
        )

    def test_load_existing_backend_type_settings(self):
        with self.settings(IEVV_JWT={
            'test_backend': {
                'ISSUER': 'test issuer',
                'AUDIENCE': 'test audience'
            }
        }):
            settings = load_settings(backend_type='test_backend')
            settings2 = DEFAULT_SETTINGS
            settings2['ISSUER'] = 'test issuer'
            settings2['AUDIENCE'] = 'test audience'
            self.assertDictEqual(
                settings,
                DEFAULT_SETTINGS
            )

    def test_load_existing_backend_type_settings_with_default_settings_override(self):
        with self.settings(IEVV_JWT={
            'default': {
                'ISSUER': 'ievv issuer'
            },
            'test_backend': {
                'AUDIENCE': 'test audience'
            }
        }):
            settings = load_settings(backend_type='test_backend')
            settings2 = DEFAULT_SETTINGS
            settings2['ISSUER'] = 'ievv issuer'
            settings2['AUDIENCE'] = 'test audience'
            self.assertDictEqual(
                settings,
                DEFAULT_SETTINGS
            )



from django.db import models

from ievv_auth.ievv_jwt_blacklist_core.models import AbstractIssuedToken, AbstractBlacklistedToken


class ScopedApiKeyIssuedToken(AbstractIssuedToken):
    scoped_api_key = models.ForeignKey(
        'ievv_api_key.ScopedAPIKey',
        on_delete=models.SET_NULL,
        null=True,
        blank=True
    )

    def blacklist_token(self) -> 'ScopedApiKeyBlacklistedToken':
        return ScopedApiKeyBlacklistedToken.objects.create(token=self)


class ScopedApiKeyBlacklistedToken(AbstractBlacklistedToken):
    token = models.OneToOneField(ScopedApiKeyIssuedToken, on_delete=models.SET_NULL, null=True, blank=True)




from ipware import get_client_ip
from rest_framework import serializers
from rest_framework.exceptions import AuthenticationFailed

from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry, get_backend_from_raw_jwt
from ievv_auth.ievv_jwt.exceptions import JWTBackendError


class RefreshTokenSerializer(serializers.Serializer):
    refresh = serializers.CharField(required=True)

    def validate(self, attrs):
        request = self.context.get("request")
        refresh_token = attrs['refresh']
        backend = get_backend_from_raw_jwt(raw_jwt=refresh_token)
        try:
            backend_instance = backend.make_instance_from_raw_jwt(raw_jwt=refresh_token, use_context=True)
        except JWTBackendError:
            raise AuthenticationFailed('Token invalid, blacklisted or expired')
        try:
            return backend_instance.refresh(token=refresh_token)
        except JWTBackendError as e:
            raise AuthenticationFailed('Token invalid, blacklisted or expired')



import base64
import json
from calendar import timegm
from unittest import mock
from unittest.mock import PropertyMock

import jwt as py_jwt
from django.test import TestCase
from django.utils import timezone
from model_bakery import baker

from ievv_auth.ievv_jwt.backends.api_key_backend import ApiKeyBackend
from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry, get_backend_from_raw_jwt
from ievv_auth.ievv_jwt.exceptions import JWTBackendError


class TestJWTBackendRegistry(TestCase):

    def setUp(self):
        registry = JWTBackendRegistry.get_instance()
        registry.set_backend(ApiKeyBackend)

    def test_sanity(self):
        registry = JWTBackendRegistry.get_instance()
        backend_class = registry.get_backend('api-key')
        self.assertEqual(backend_class, ApiKeyBackend)

    def test_unknown_backend(self):
        registry = JWTBackendRegistry.get_instance()
        backend_class = registry.get_backend('unknown')
        self.assertIsNone(backend_class)

    def test_get_backend_from_raw_jwt(self):
        api_key = baker.make(
            'ievv_api_key.ScopedApiKey',
            base_jwt_payload={
                'scope': ['read', 'write']
            }
        )
        backend = ApiKeyBackend()
        backend.set_context(api_key_instance=api_key)
        jwt = backend.encode_access_token()
        backend_class = get_backend_from_raw_jwt(raw_jwt=jwt)
        self.assertEqual(backend_class, ApiKeyBackend)

    def test_get_backend_from_raw_jwt_unknown_backend(self):
        backend_class = get_backend_from_raw_jwt(raw_jwt='eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJhcGlfa2V5X2lkIjo2LCJleHAiOjE1NzkwODI2MjEsImlhdCI6MTU3OTA4MDgyMSwianRpIjoiZTRkMTVkZTQwZjg1NDk1Mjg2MGNkYTI3N2Y0NDIwYWIiLCJqd3RfYmFja2VuZF9uYW1lIjoidW5rbm93biJ9.9PM35De6yg_IaTvdLgBzo2F5xG9ywioIthTOK637iO4')
        self.assertIsNone(backend_class)



from ievv_auth.ievv_jwt.api.serializers.api_key_serializers import ApiKeyObtainJWTSerializer
from ievv_auth.ievv_jwt.api.views.base_view import JWTBaseView


class APIKeyObtainJWTAccessTokenView(JWTBaseView):

    serializer_class = ApiKeyObtainJWTSerializer



from ipware import get_client_ip
from rest_framework import serializers
from rest_framework.exceptions import AuthenticationFailed

from django.apps import apps
from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry
from ievv_auth.ievv_jwt.exceptions import JWTBackendError


class ApiKeyObtainJWTSerializer(serializers.Serializer):
    api_key = serializers.CharField()

    def validate(self, attrs):
        request = self.context.get("request")
        client_ip, is_routable = get_client_ip(request)
        ScopedAPIKey = apps.get_model(app_label='ievv_api_key', model_name='ScopedAPIKey')
        is_valid, instance = ScopedAPIKey.objects.is_valid_with_logging(
            api_key=attrs['api_key'],
            extra={
                'ip': client_ip
            }
        )
        if not is_valid or not instance:
            raise AuthenticationFailed('Api key has expired or is invalid')
        jwt_backend_class = JWTBackendRegistry.get_instance().get_backend(instance.jwt_backend_name)
        if not jwt_backend_class:
            raise AuthenticationFailed('Unknown jwt backend could not authenticate')
        backend = jwt_backend_class()
        backend.set_context(api_key_instance=instance)
        return backend.make_authenticate_success_response()



from django import test
from django.utils import timezone

from ievv_auth.ievv_api_key.models import ScopedAPIKey
from ievv_auth.ievv_jwt.api.views import APIKeyObtainJWTAccessTokenView
from ievv_auth.ievv_jwt.backends.api_key_backend import ApiKeyBackend
from ievv_auth.ievv_jwt.tests.test_api import api_test_mixin


class TestAPIKeyObtainJWTAccessTokenView(test.TestCase, api_test_mixin.ApiTestMixin):
    apiview_class = APIKeyObtainJWTAccessTokenView

    def setUp(self):
        from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry
        from ievv_auth.ievv_jwt.backends.api_key_backend import ApiKeyBackend
        registry = JWTBackendRegistry.get_instance()
        registry.set_backend(ApiKeyBackend)

    def test_no_api_key(self):
        response = self.make_post_request()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data.get('api_key')[0], 'This field is required.')

    def test_wrong_api_key(self):
        response = self.make_post_request(data={'api_key': '12ret'})
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response.data.get('detail'), 'Api key has expired or is invalid')

    def test_unknown_backend_key(self):
        api_key, instance = ScopedAPIKey.objects.create_key(
            name='test',
            jwt_backend_name='crazy',
        )
        response = self.make_post_request(data={'api_key': api_key})
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response.data.get('detail'), 'Unknown jwt backend could not authenticate')

    def test_api_key_revoked(self):
        api_key, instance = ScopedAPIKey.objects.create_key(
            name='test',
            jwt_backend_name='crazy',
        )
        instance.revoked = True
        instance.save()
        response = self.make_post_request(data={'api_key': api_key})
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response.data.get('detail'), 'Api key has expired or is invalid')

    def test_api_key_expired(self):
        api_key, instance = ScopedAPIKey.objects.create_key(
            name='test',
            jwt_backend_name='crazy',
        )
        instance.expiration_datetime = timezone.now()
        instance.save()
        response = self.make_post_request(data={'api_key': api_key})
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response.data.get('detail'), 'Api key has expired or is invalid')

    def test_api_key_ok_api_key_backend(self):
        api_key, instance = ScopedAPIKey.objects.create_key(
            name='test',
            jwt_backend_name='api-key',
            base_jwt_payload={
                'scope': 'cool'
            }
        )
        response = self.make_post_request(data={'api_key': api_key})
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.data.get('access'))
        payload = ApiKeyBackend().decode(token=response.data.get('access'), verify=True)
        self.assertEqual(payload['scope'], 'cool')
        self.assertEqual(payload['api_key_id'], instance.id)
        self.assertEqual(payload['jwt_backend_name'], 'api-key')



import json
import os

"""
A simple Django settings module proxy that lets us configure Django
using the DJANGOENV environment variable.

Example (running tests)::

    $ DJANGOENV=test python manage.py test

Defaults to the ``develop`` enviroment, so developers can use ``python
manage.py`` without anything extra during development.
"""

DJANGOENV = os.environ.get('DJANGOENV', 'develop')


def _load_develop_environment_from_file():
    if os.path.exists('develop-environment.json'):
        environment_dict = json.loads(open('develop-environment.json', 'r').read())
        for key, value in environment_dict.items():
            if key.upper() == key:
                os.environ[key] = value

if DJANGOENV == 'develop':  # Used for local development
    _load_develop_environment_from_file()
    from ievv_auth.project.develop.develop_settings import *
elif DJANGOENV == 'test':  # Used when running the Django tests locally
    _load_develop_environment_from_file()
    from ievv_auth.project.develop.test_settings import *
# elif DJANGOENV == 'production':  # Used in production
#     from ievv_auth.project.production.settings import *
else:
    raise ValueError('Invalid value for the DJANGOENV environment variable: {}'.format(DJANGOENV))



import typing as t

from django.contrib.auth import get_user_model
from django.apps import apps

from ievv_auth.ievv_jwt.backends.base_backend import AbstractBackend
from ievv_auth.ievv_jwt.exceptions import JWTBackendError

UserModel = get_user_model()


if t.TYPE_CHECKING:
    from ievv_auth.ievv_jwt_blacklist_user.models import UserBlacklistedToken, UserIssuedToken


class UserAuthBackend(AbstractBackend):
    user_instance: UserModel = None

    @classmethod
    def get_backend_name(cls):
        return 'user-auth'

    def set_context(self, user_instance: UserModel = None, *args, **kwargs):
        self.user_instance = user_instance

    @property
    def issued_token_model(self) -> t.Type['UserIssuedToken']:
        return apps.get_model(app_label='ievv_jwt_blacklist_user', model_name='UserIssuedToken')

    @property
    def blacklisted_token_model(self) -> t.Type['UserBlacklistedToken']:
        return apps.get_model(app_label='ievv_jwt_blacklist_user', model_name='UserBlacklistedToken')

    def create_issued_token(self, token, payload, issued_at, expires_at, jti) -> 'UserBlacklistedToken':
        return self.issued_token_model.objects.create(
            issued_at=issued_at,
            expires_at=expires_at,
            token=token,
            jti=jti,
            token_payload=payload,
            user=self.user_instance
        )

    def make_access_token_payload(self) -> dict:
        if self.user_instance is None:
            raise JWTBackendError('Missing context "user_instance"')
        return {
            'user_id': self.user_instance.id
        }

    def make_refresh_token_payload(self) -> dict:
        if self.user_instance is None:
            raise JWTBackendError('Missing context "user_instance"')
        return {
            'user_id': self.user_instance.id
        }

    @classmethod
    def make_instance_from_raw_jwt(cls, raw_jwt, use_context=False, *args, **kwargs):
        instance = cls()
        if use_context:
            payload = instance.decode(raw_jwt)
            if 'user_id' not in payload:
                raise JWTBackendError('JWT payload missing key "user_id"')
            try:
                user_instance = UserModel.objects.get(id=payload['user_id'])
                if not user_instance.is_active:
                    raise JWTBackendError('Inactive User')
                instance.set_context(user_instance=user_instance)
            except UserModel.DoesNotExist:
                raise JWTBackendError(f'No user with id "{payload["user_id"]}" found')
        return instance



import base64
import json
from calendar import timegm
from unittest import mock
from unittest.mock import PropertyMock

import jwt as py_jwt
from django.test import TestCase
from django.utils import timezone
from model_bakery import baker
from django.conf import settings

from ievv_auth.ievv_jwt.backends.user_auth_backend import UserAuthBackend
from ievv_auth.ievv_jwt.exceptions import JWTBackendError


class TestUserAuthBackend(TestCase):

    def test_access_token_sanity(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_access_token(base_payload={
            'scope': ['read', 'write']
        })
        decoded = backend.decode(jwt)
        self.assertIn('exp', decoded)
        self.assertIn('iat', decoded)
        self.assertIn('jti', decoded)
        self.assertEqual(decoded['user_id'], user.id)
        self.assertEqual(decoded['scope'], ['read', 'write'])
        self.assertEqual(decoded[backend.settings['TOKEN_TYPE_CLAIM']], 'access')

    def test_refresh_token_sanity(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_refresh_token(base_payload={
            'scope': ['read', 'write']
        })
        decoded = backend.decode(jwt)
        self.assertIn('exp', decoded)
        self.assertIn('iat', decoded)
        self.assertIn('jti', decoded)
        self.assertEqual(decoded['user_id'], user.id)
        self.assertEqual(decoded['scope'], ['read', 'write'])
        self.assertEqual(decoded[backend.settings['TOKEN_TYPE_CLAIM']], 'refresh')

    def test_access_fields_which_is_not_overridable_is_not_changed(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_access_token(base_payload={
            'exp': 123,
            'iat': 123,
            'jti': 123,
        })
        decoded = backend.decode(jwt)
        self.assertIn('exp', decoded)
        self.assertNotEqual(decoded['exp'], 123)
        self.assertIn('iat', decoded)
        self.assertNotEqual(decoded['iat'], 123)
        self.assertIn('jti', decoded)
        self.assertNotEqual(decoded['jti'], 123)
        self.assertEqual(decoded['user_id'], user.id)

    def test_refresh_fields_which_is_not_overridable_is_not_changed(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_refresh_token(base_payload={
            'exp': 123,
            'iat': 123,
            'jti': 123,
        })
        decoded = backend.decode(jwt)
        self.assertIn('exp', decoded)
        self.assertNotEqual(decoded['exp'], 123)
        self.assertIn('iat', decoded)
        self.assertNotEqual(decoded['iat'], 123)
        self.assertIn('jti', decoded)
        self.assertNotEqual(decoded['jti'], 123)
        self.assertEqual(decoded['user_id'], user.id)

    def test_access_verify_intercepted_payload_extend_expiration(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_access_token()
        [header, _, secret] = jwt.split('.')
        decoded = backend.decode(jwt)
        decoded['exp'] = timegm((timezone.now() + timezone.timedelta(weeks=200)).utctimetuple())
        payload = base64.urlsafe_b64encode(
            json.dumps(
                decoded,
                separators=(',', ':')
            ).encode('utf-8')
        ).decode('utf-8')
        new_jwt = f'{header}.{payload}.{secret}'
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_refresh_intercepted_payload_extend_expiration(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_refresh_token()
        [header, _, secret] = jwt.split('.')
        decoded = backend.decode(jwt)
        decoded['exp'] = timegm((timezone.now() + timezone.timedelta(weeks=200)).utctimetuple())
        payload = base64.urlsafe_b64encode(
            json.dumps(
                decoded,
                separators=(',', ':')
            ).encode('utf-8')
        ).decode('utf-8')
        new_jwt = f'{header}.{payload}.{secret}'
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_access_verify_intercepted_payload_added_additional_scope(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_access_token()
        [header, _, secret] = jwt.split('.')
        decoded = backend.decode(jwt)
        decoded['scope'] = 'admin'
        payload = base64.urlsafe_b64encode(
            json.dumps(
                decoded,
                separators=(',', ':')
            ).encode('utf-8')
        ).decode('utf-8')
        new_jwt = f'{header}.{payload}.{secret}'
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_refresh_verify_intercepted_payload_added_additional_scope(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_refresh_token()
        [header, _, secret] = jwt.split('.')
        decoded = backend.decode(jwt)
        decoded['scope'] = 'admin'
        payload = base64.urlsafe_b64encode(
            json.dumps(
                decoded,
                separators=(',', ':')
            ).encode('utf-8')
        ).decode('utf-8')
        new_jwt = f'{header}.{payload}.{secret}'
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_sign_access_jwt_with_another_secret(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_access_token()
        decoded = backend.decode(jwt)
        new_jwt = py_jwt.encode(payload=decoded, key='asdxxxxxxxxxxxxxxxxxxxxxxxxxxx')
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)

    def test_sign_refresh_jwt_with_another_secret(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_refresh_token()
        decoded = backend.decode(jwt)
        new_jwt = py_jwt.encode(payload=decoded, key='asdxxxxxxxxxxxxxxxxxxxxxxxxxxx')
        with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
            backend.decode(new_jwt, verify=True)


    def test_access_token_has_expired(self):
        with self.settings(IEVV_JWT={
            'default': {
                'ACCESS_TOKEN_LIFETIME': timezone.timedelta(minutes=0),
            }
        }):
            with mock.patch(
                    'ievv_auth.ievv_jwt.backends.user_auth_backend.UserAuthBackend.access_token_expiration',
                    new_callable=PropertyMock,
                    return_value=timezone.now() - timezone.timedelta(days=1)):
                user = baker.make(settings.AUTH_USER_MODEL)
                backend = UserAuthBackend()
                backend.set_context(user_instance=user)
                jwt = backend.encode_access_token()
                with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
                    backend.decode(jwt, verify=True)

    def test_refresh_token_has_expired(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(minutes=0),
            }
        }):
            with mock.patch(
                    'ievv_auth.ievv_jwt.backends.user_auth_backend.UserAuthBackend.refresh_token_expiration',
                    new_callable=PropertyMock,
                    return_value=timezone.now() - timezone.timedelta(days=1)):
                user = baker.make(settings.AUTH_USER_MODEL)
                backend = UserAuthBackend()
                backend.set_context(user_instance=user)
                jwt = backend.encode_refresh_token()
                with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
                    backend.decode(jwt, verify=True)

    def test_make_instance_from_raw_jwt(self):
        user = baker.make(settings.AUTH_USER_MODEL)
        backend = UserAuthBackend()
        backend.set_context(user_instance=user)
        jwt = backend.encode_access_token()
        backend_instance = UserAuthBackend.make_instance_from_raw_jwt(raw_jwt=jwt)
        self.assertIsInstance(backend_instance, UserAuthBackend)

    def test_make_authenticate_success_access_token_only_response(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': None,
            }
        }):
            with mock.patch(
                'ievv_auth.ievv_jwt.backends.user_auth_backend.UserAuthBackend.encode_access_token',
                return_value='test'
            ):
                user = baker.make(settings.AUTH_USER_MODEL)
                backend = UserAuthBackend()
                backend.set_context(user_instance=user)
                self.assertDictEqual(
                    backend.make_authenticate_success_response(),
                    {'access': 'test'}
                )

    def test_make_authenticate_success_access_refresh_response(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
            }
        }):
            with mock.patch(
                'ievv_auth.ievv_jwt.backends.user_auth_backend.UserAuthBackend.encode_access_token',
                return_value='access token'
            ):
                with mock.patch(
                        'ievv_auth.ievv_jwt.backends.user_auth_backend.UserAuthBackend.encode_refresh_token',
                        return_value='refresh token'
                ):
                    user = baker.make(settings.AUTH_USER_MODEL)
                    backend = UserAuthBackend()
                    backend.set_context(user_instance=user)
                    self.assertDictEqual(
                        backend.make_authenticate_success_response(),
                        {'access': 'access token', 'refresh': 'refresh token'}
                    )

    def test_refresh_token_wrong_token_type(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            with self.assertRaisesMessage(JWTBackendError, 'Token is not a refresh token'):
                new_token_pair = backend.refresh(token=token_pair['access'])

    def test_make_authenticate_blacklist_not_installed_in_apps(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_IEVV_JWT):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            self.assertIn('access', token_pair)
            self.assertIn('refresh', token_pair)
            with self.settings(INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
                self.assertEqual(backend.issued_token_model.objects.all().count(), 0)

    def test_make_authenticate_blacklist_in_installed_apps_but_not_enabled(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': False
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            self.assertIn('access', token_pair)
            self.assertIn('refresh', token_pair)
            self.assertEqual(backend.issued_token_model.objects.all().count(), 0)

    def test_make_authenticate_blacklist_in_use(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            self.assertIn('access', token_pair)
            self.assertIn('refresh', token_pair)
            refresh_jti = backend.decode(token_pair['refresh'])[backend.settings['JTI_CLAIM']]
            IssuedTokenModel = backend.issued_token_model
            self.assertEqual(IssuedTokenModel.objects.filter(jti=refresh_jti).count(), 1)

    def test_refresh_toke_blacklist_not_in_installed_apps(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_API_KEY):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            new_token_pair = backend.refresh(token=token_pair['refresh'])
            self.assertIn('access', new_token_pair)
            self.assertIn('refresh', new_token_pair)
            with self.settings(INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
                self.assertEqual(backend.issued_token_model.objects.all().count(), 0)
                self.assertEqual(backend.blacklisted_token_model.objects.all().count(), 0)
            self.assertNotEqual(token_pair['access'], new_token_pair['access'])
            self.assertNotEqual(token_pair['refresh'], new_token_pair['refresh'])

    def test_refresh_token_blacklist_is_in_installed_apps_but_blacklist_is_disabled(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': False
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            new_token_pair = backend.refresh(token=token_pair['refresh'])
            self.assertIn('access', new_token_pair)
            self.assertIn('refresh', new_token_pair)
            with self.settings(INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
                self.assertEqual(backend.issued_token_model.objects.all().count(), 0)
                self.assertEqual(backend.blacklisted_token_model.objects.all().count(), 0)
            self.assertNotEqual(token_pair['access'], new_token_pair['access'])
            self.assertNotEqual(token_pair['refresh'], new_token_pair['refresh'])

    def test_refresh_token_is_blacklisted(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            refresh_jti = backend.decode(token_pair['refresh'])[backend.settings['JTI_CLAIM']]
            IssuedTokenModel = backend.issued_token_model
            BlacklistedTokenModel = backend.blacklisted_token_model
            IssuedTokenModel.objects.get(jti=refresh_jti).blacklist_token()
            self.assertEqual(BlacklistedTokenModel.objects.filter(token__jti=refresh_jti).count(), 1)
            with self.assertRaisesMessage(JWTBackendError, 'Token is not valid'):
                new_token_pair = backend.refresh(token=token_pair['refresh'])

    def test_refresh_token_blacklist_on_rotation(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            refresh_jti = backend.decode(token_pair['refresh'])[backend.settings['JTI_CLAIM']]
            IssuedTokenModel = backend.issued_token_model
            BlacklistedTokenModel = backend.blacklisted_token_model
            new_token_pair = backend.refresh(token=token_pair['refresh'])
            self.assertEqual(BlacklistedTokenModel.objects.filter(token__jti=refresh_jti).count(), 1)

    def test_refresh_token_blacklist_on_rotation_try_to_use_old(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            refresh_jti = backend.decode(token_pair['refresh'])[backend.settings['JTI_CLAIM']]
            IssuedTokenModel = backend.issued_token_model
            BlacklistedTokenModel = backend.blacklisted_token_model
            new_token_pair = backend.refresh(token=token_pair['refresh'])
            self.assertEqual(BlacklistedTokenModel.objects.filter(token__jti=refresh_jti).count(), 1)
            with self.assertRaisesMessage(JWTBackendError, 'Token is not valid'):
                new_token_pair = backend.refresh(token=token_pair['refresh'])

    def test_refresh_request_token_has_expired(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True
            }
        }, INSTALLED_APPS=settings.INSTALLED_APPS_USER_BLACKLIST):
            with mock.patch(
                    'ievv_auth.ievv_jwt.backends.user_auth_backend.UserAuthBackend.refresh_token_expiration',
                    new_callable=PropertyMock,
                    return_value=timezone.now() - timezone.timedelta(days=1)):
                user = baker.make(settings.AUTH_USER_MODEL)
                backend = UserAuthBackend()
                backend.set_context(user_instance=user)
                token_pair = backend.make_authenticate_success_response()
                refresh_jti = py_jwt.decode(
                    jwt=token_pair['refresh'],
                    verify=False,
                    options={'verify_signature': False}
                )[backend.settings['JTI_CLAIM']]
                IssuedTokenModel = backend.issued_token_model
                with self.assertRaisesMessage(JWTBackendError, 'Token is invalid or expired'):
                    backend.refresh(token=token_pair['refresh'])



from django.conf import settings
from model_bakery import baker
from rest_framework.test import APIRequestFactory, force_authenticate


class ApiTestMixin:
    """
    Mixin class for API tests.

    Can be used for ViewSets too. Just override :meth:`.get_as_view_kwargs` - se example in the docs
    for that method.
    """
    apiview_class = None

    def get_default_requestuser(self):
        return None

    def make_user(self, email='user@example.com', **kwargs):
        return baker.make(settings.AUTH_USER_MODEL, email=email, **kwargs)

    def make_superuser(self, email='super@example.com', **kwargs):
        return self.make_user(email=email, is_superuser=True, **kwargs)

    def get_as_view_kwargs(self):
        """
        The kwargs for the ``as_view()``-method of the API view class.

        If you are writing tests for a ViewSet, you have to override this
        to define what action you are testing (list, retrieve, update, ...), like this::

            def get_as_view_kwargs(self):
                return {
                    'actions: {
                        'get': 'list'
                    }
                }
        """
        return {}

    def add_authenticated_user_to_request(self, request, requestuser):
        if requestuser:
            force_authenticate(request, requestuser)

    def make_request(self, method, viewkwargs=None, api_url='/test/', data=None, requestuser=None):
        factory = APIRequestFactory()
        request = getattr(factory, method)(api_url, format='json', data=data)
        viewkwargs = viewkwargs or {}
        if requestuser:
            request.user = requestuser or self.get_default_requestuser()
        response = self.apiview_class.as_view(**self.get_as_view_kwargs())(request, **viewkwargs)
        response.render()
        return response

    def make_get_request(self, viewkwargs=None, api_url='/test/', data=None, requestuser=None):
        return self.make_request(method='get', viewkwargs=viewkwargs,
                                 api_url=api_url, data=data,
                                 requestuser=requestuser)

    def make_post_request(self, viewkwargs=None, api_url='/test/', data=None, requestuser=None):
        return self.make_request(method='post', viewkwargs=viewkwargs,
                                 api_url=api_url, data=data,
                                 requestuser=requestuser)

    def make_put_request(self, viewkwargs=None, api_url='/test/', data=None, requestuser=None):
        return self.make_request(method='put', viewkwargs=viewkwargs,
                                 api_url=api_url, data=data,
                                 requestuser=requestuser)

    def make_delete_request(self, viewkwargs=None, api_url='/test/', data=None, requestuser=None):
        return self.make_request(method='delete', viewkwargs=viewkwargs,
                                 api_url=api_url, data=data,
                                 requestuser=requestuser)



import datetime
from json import JSONEncoder


class DateTimeEncoder(JSONEncoder):
    # Override the default method
    def default(self, obj):
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()



import typing as t

import jwt
from ievv_opensource.utils.singleton import Singleton

if t.TYPE_CHECKING:
    from .base_backend import AbstractBackend


class JWTBackendRegistry(Singleton):

    def __init__(self):
        super(JWTBackendRegistry, self).__init__()
        self._backend_map = {}

    def set_backend(self, backend: t.Type['AbstractBackend']):
        self._backend_map[backend.get_backend_name()] = backend

    def get_backend(self, backend_name) -> t.Type['AbstractBackend']:
        return self._backend_map.get(backend_name, None)

    def get_backend_choices(self):
        choices = []
        for key in self._backend_map:
            choices.append((key, key))
        return choices


def get_backend_from_raw_jwt(raw_jwt) -> t.Type['AbstractBackend']:
    """
    Gets the backend from raw_jwt,
    this step does not verify the jwt, it should happen later,
    the only purpose is to get the backend class
    Args:
        raw_jwt: raw jwt string

    Returns:
        instance of AbstractBackend
    """
    payload = jwt.decode(jwt=raw_jwt, verify=False, options={'verify_signature': False})
    return JWTBackendRegistry.get_instance().get_backend(payload['jwt_backend_name'])



import typing as t

from rest_framework import serializers
from rest_framework.exceptions import AuthenticationFailed
from django.contrib.auth import get_user_model, authenticate
from django.contrib.auth.models import update_last_login

from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry
from ievv_auth.ievv_jwt.exceptions import JWTBackendError

User = get_user_model()
username_field = User.USERNAME_FIELD


class PasswordField(serializers.CharField):
    def __init__(self, **kwargs) -> None:
        kwargs.setdefault("style", {})
        kwargs["style"]["input_type"] = "password"
        kwargs["write_only"] = True
        super().__init__(**kwargs)


class UserAuthObtainJWTSerializer(serializers.Serializer):
    jwt_backend_name = 'default'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.fields[username_field] = serializers.CharField(write_only=True, required=True)
        self.fields['password'] = PasswordField(required=True)

    def validate(self, attrs):
        authenticate_kwargs = {
            username_field: attrs[username_field],
            'password': attrs['password'],
            'request': self.context.get("request")
        }
        user_instance = authenticate(**authenticate_kwargs)
        if user_instance is None or not user_instance.is_active:
            raise AuthenticationFailed()
        jwt_backend_class = JWTBackendRegistry.get_instance().get_backend(self.jwt_backend_name)
        if not jwt_backend_class:
            raise AuthenticationFailed('Unknown jwt backend could not authenticate')
        backend = jwt_backend_class()
        if backend.update_last_login:
            update_last_login(None, user_instance)
        backend.set_context(user_instance=user_instance)
        return backend.make_authenticate_success_response()


def get_user_auth_obtain_jwt_serializer(backend_name='user-auth') -> t.Type[UserAuthObtainJWTSerializer]:
    class _UserAuthObtainJWTSerializer(UserAuthObtainJWTSerializer):
        jwt_backend_name = backend_name
    return _UserAuthObtainJWTSerializer



from django.db import models


class AbstractIssuedToken(models.Model):

    #: JTI Claim
    jti = models.CharField(
        unique=True,
        max_length=255,
        editable=False
    )

    #: JWT token
    token = models.TextField(null=False, blank=False, editable=False)

    #: JWT token payload
    token_payload = models.JSONField(null=False, blank=False, editable=False)

    #: Token issued at
    issued_at = models.DateTimeField(null=False, blank=False, editable=False)

    #: Token expires at
    expires_at = models.DateTimeField(null=False, blank=False, editable=False)

    class Meta:
        abstract = True

    def blacklist_token(self) -> 'AbstractBlacklistedToken':
        raise NotImplementedError('Implement blacklist_token method')


class AbstractBlacklistedToken(models.Model):
    blacklisted_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True



from rest_framework import HTTP_HEADER_ENCODING, authentication
from rest_framework.exceptions import AuthenticationFailed
from django.utils.translation import gettext_lazy

from ievv_auth.ievv_jwt.backends.backend_registry import get_backend_from_raw_jwt
from ievv_auth.ievv_jwt.exceptions import JWTBackendError
from ievv_auth.ievv_jwt.utils.load_settings import load_global_settings

ievv_jwt_global_settings = load_global_settings()
AUTH_HEADER_TYPES = ievv_jwt_global_settings['AUTH_HEADER_TYPES']

AUTH_HEADER_TYPE_BYTES = set(
    h.encode(HTTP_HEADER_ENCODING)
    for h in AUTH_HEADER_TYPES
)


class JWTAuthentication(authentication.BaseAuthentication):
    www_authenticate_realm = 'api'

    def authenticate(self, request):
        header = self.get_header(request)
        if header is None:
            return None

        raw_token = self.get_raw_token(header)
        if raw_token is None:
            return None

        backend_class = get_backend_from_raw_jwt(raw_token)
        if backend_class:
            backend = backend_class.make_instance_from_raw_jwt(raw_token)
            try:
                payload = backend.decode(token=raw_token, verify=True)
                return None, payload
            except JWTBackendError:
                raise AuthenticationFailed('Token verification failed, token is invalid or has expired')
        raise AuthenticationFailed('No matching backend found for token')

    def get_header(self, request):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        header = request.META.get('HTTP_AUTHORIZATION')

        if isinstance(header, str):
            # Work around django test client oddness
            header = header.encode(HTTP_HEADER_ENCODING)

        return header

    def authenticate_header(self, request):

        return '{0} realm="{1}"'.format(
            AUTH_HEADER_TYPES[0],
            self.www_authenticate_realm,
        )

    def get_raw_token(self, header):
        """
        Extracts an unvalidated JSON web token from the given "Authorization"
        header value.
        """
        parts = header.split()

        if len(parts) == 0:
            # Empty AUTHORIZATION header sent
            return None

        if parts[0] not in AUTH_HEADER_TYPE_BYTES:
            # Assume the header does not contain a JSON web token
            return None

        if len(parts) != 2:
            raise AuthenticationFailed(
                gettext_lazy('Authorization header must contain two space-delimited values'),
                code='bad_authorization_header',
            )

        return parts[1]
